@@ -32,6 +32,15 @@
 #define DEBUG(x)
 #endif
 
+#include "gtksourceview-i18n.h"
+#include "gtksourcebuffer.h"
+#include "gtksourcelanguage.h"
+#include "gtksourcelanguage-private.h"
+#include "gtksourcecontextengine.h"
+
+#include <glib.h>
+#include <glib/gstdio.h>
+
 #include <string.h>
 #include <fcntl.h>
 #ifdef HAVE_UNISTD_H
@@ -41,13 +50,6 @@
 #include <io.h>
 #endif
 #include <libxml/xmlreader.h>
-#include <glib/gstdio.h>
-#include "gtksourceview-i18n.h"
-#include "gtksourcebuffer.h"
-#include "gtksourcelanguage.h"
-#include "gtksourcelanguage-private.h"
-#include "gtksourcecontextengine.h"
-#include <glib/gregex.h>
 
 #define PARSER_ERROR (parser_error_quark ())
 #define ATTR_NO_STYLE ""
@@ -67,6 +69,7 @@ struct _ParserState
 {
 	/* The args passed to _file_parse_version2() */
 	xmlTextReader *reader;
+	char *filename;
 	GtkSourceLanguage *language;
 	GtkSourceContextData *ctx_data;
 
@@ -130,6 +133,7 @@ static ParserState *parser_state_new           (GtkSourceLanguage      *language
                                                 GHashTable             *styles_mapping,
 						GQueue                 *replacements,
                                                 xmlTextReader          *reader,
+						const char             *filename,
                                                 GHashTable             *loaded_lang_ids);
 static void       parser_state_destroy         (ParserState *parser_state);
 
@@ -151,6 +155,7 @@ static gboolean   create_definition            (ParserState *parser_state,
                                                 gchar *id,
                                                 gchar *parent_id,
                                                 gchar *style,
+                                                GSList *context_classes,
 						GError **error);
 
 static void       handle_context_element       (ParserState *parser_state);
@@ -249,14 +254,14 @@ static GRegexCompileFlags
 get_regex_flags (xmlNode             *node,
 		 GRegexCompileFlags flags)
 {
-	xmlAttr *attribute;
+	xmlAttr *attr;
 
-	for (attribute = node->properties; attribute != NULL; attribute = attribute->next)
+	for (attr = node->properties; attr != NULL; attr = attr->next)
 	{
-		g_return_val_if_fail (attribute->children != NULL, flags);
+		g_return_val_if_fail (attr->children != NULL, flags);
 
-		flags = update_regex_flags (flags, attribute->name,
-					    attribute->children->content);
+		flags = update_regex_flags (flags, attr->name,
+					    attr->children->content);
 	}
 
 	return flags;
@@ -301,11 +306,56 @@ get_context_flags (ParserState *parser_state)
 	return flags;
 }
 
+static GSList *
+add_classes (GSList      *list,
+             gchar const *classes,
+             gboolean     enabled)
+{
+	gchar **parts;
+	gchar **ptr;
+	GSList *newlist = NULL;
+
+	if (classes == NULL)
+	{
+		return list;
+	}
+
+	parts = ptr = g_strsplit (classes, " ", -1);
+
+	while (*ptr)
+	{
+		GtkSourceContextClass *ctx = gtk_source_context_class_new (*ptr, enabled);
+		newlist = g_slist_prepend (newlist, ctx);
+
+		++ptr;
+	}
+
+	g_strfreev (parts);
+	return g_slist_concat (list, g_slist_reverse (newlist));
+}
+
+static GSList *
+parse_classes (ParserState *parser_state)
+{
+	GSList *ret = NULL;
+	xmlChar *en = xmlTextReaderGetAttribute (parser_state->reader, BAD_CAST "class");
+	xmlChar *dis = xmlTextReaderGetAttribute (parser_state->reader, BAD_CAST "class-disabled");
+
+	ret = add_classes (ret, (gchar const *)en, TRUE);
+	ret = add_classes (ret, (gchar const *)dis, FALSE);
+
+	xmlFree (en);
+	xmlFree (dis);
+
+	return ret;
+}
+
 static gboolean
 create_definition (ParserState *parser_state,
 		   gchar       *id,
 		   gchar       *parent_id,
 		   gchar       *style,
+		   GSList      *context_classes,
 		   GError     **error)
 {
 	gchar *match = NULL, *start = NULL, *end = NULL;
@@ -466,6 +516,7 @@ create_definition (ParserState *parser_state,
 	}
 
 	if (tmp_error == NULL)
+	{
 		_gtk_source_context_data_define_context (parser_state->ctx_data,
 							 id,
 							 parent_id,
@@ -473,8 +524,10 @@ create_definition (ParserState *parser_state,
 							 start,
 							 end,
 							 style,
+							 context_classes,
 							 flags,
 							 &tmp_error);
+	}
 
 	g_free (match);
 	g_free (start);
@@ -606,11 +659,12 @@ add_ref (ParserState               *parser_state,
 }
 
 static gboolean
-create_sub_pattern (ParserState *parser_state,
-                    gchar *id,
-                    gchar *sub_pattern,
-                    gchar *style,
-                    GError **error)
+create_sub_pattern (ParserState  *parser_state,
+                    gchar        *id,
+                    gchar        *sub_pattern,
+                    gchar        *style,
+                    GSList       *context_classes,
+                    GError      **error)
 {
 	gchar *container_id;
 	xmlChar *where;
@@ -632,6 +686,7 @@ create_sub_pattern (ParserState *parser_state,
 						  sub_pattern,
 						  (gchar*) where,
 						  style,
+						  context_classes,
 						  &tmp_error);
 
 	xmlFree (where);
@@ -654,6 +709,7 @@ handle_context_element (ParserState *parser_state)
 	gboolean success;
 	gboolean ignore_style = FALSE;
 	GtkSourceContextRefOptions options = 0;
+	GSList *context_classes;
 
 	GError *tmp_error = NULL;
 
@@ -693,16 +749,19 @@ handle_context_element (ParserState *parser_state)
 		options |= GTK_SOURCE_CONTEXT_IGNORE_STYLE;
 
 		if (style_ref != NULL)
-			g_warning ("style-ref and ignore-style used simultaneously");
+			g_warning ("in file %s: style-ref and ignore-style used simultaneously",
+				   parser_state->filename);
 	}
 
 	/* XXX */
 	if (!ignore_style && style_ref != NULL &&
 	    g_hash_table_lookup (parser_state->styles_mapping, style_ref) == NULL)
 	{
-		g_warning ("style '%s' not defined", style_ref);
+		g_warning ("in file %s: style '%s' not defined", parser_state->filename, style_ref);
 	}
 
+	context_classes = parse_classes (parser_state);
+
 	if (ref != NULL)
 	{
 		tmp = xmlTextReaderGetAttribute (parser_state->reader, BAD_CAST "original");
@@ -713,7 +772,11 @@ handle_context_element (ParserState *parser_state)
 		if (style_ref != NULL)
 			options |= GTK_SOURCE_CONTEXT_OVERRIDE_STYLE;
 
-		add_ref (parser_state, (gchar*) ref, options, style_ref, &tmp_error);
+		add_ref (parser_state,
+		         (gchar*) ref,
+		         options,
+		         style_ref,
+		         &tmp_error);
 	}
 	else
 	{
@@ -738,9 +801,12 @@ handle_context_element (ParserState *parser_state)
 		{
 			if (sub_pattern != NULL)
 			{
-				create_sub_pattern (parser_state, id,
-						(gchar *)sub_pattern, style_ref,
-						&tmp_error);
+				create_sub_pattern (parser_state,
+				                    id,
+				                    (gchar *)sub_pattern,
+				                    style_ref,
+				                    context_classes,
+				                    &tmp_error);
 			}
 			else
 			{
@@ -758,11 +824,16 @@ handle_context_element (ParserState *parser_state)
 											   NULL,
 											   NULL,
 											   NULL,
+											   NULL,
 											   0,
 											   &tmp_error);
 				else
-					success = create_definition (parser_state, id, parent_id,
-								     style_ref, &tmp_error);
+					success = create_definition (parser_state,
+					                             id,
+					                             parent_id,
+					                             style_ref,
+					                             context_classes,
+					                             &tmp_error);
 
 				if (success && !is_empty)
 				{
@@ -778,6 +849,9 @@ handle_context_element (ParserState *parser_state)
 		g_free (id);
 	}
 
+	g_slist_foreach (context_classes, (GFunc)gtk_source_context_class_free, NULL);
+	g_slist_free (context_classes);
+
 	g_free (style_ref);
 	xmlFree (sub_pattern);
 	xmlFree (ref);
@@ -1236,11 +1310,13 @@ handle_define_regex_element (ParserState *parser_state)
 		DEBUG (g_message ("defined regex %s: \"%s\"", id, (gchar *)regex));
 		g_hash_table_insert (parser_state->defined_regexes, id, expanded_regex);
 	}
+	else
+	{
+		g_propagate_error (&parser_state->error, tmp_error);
+		g_free (id);
+	}
 
 	xmlFree (regex);
-
-	if (tmp_error != NULL)
-		g_propagate_error (&parser_state->error, tmp_error);
 }
 
 static void
@@ -1356,6 +1432,12 @@ parse_style (ParserState *parser_state)
 	DEBUG (g_message ("style %s (%s) to be mapped to '%s'",
 			  name, id, map_to ? (char*) map_to : "(null)"));
 
+	if (map_to != NULL &&
+	    g_hash_table_lookup (parser_state->styles_mapping, map_to) == NULL)
+	{
+		g_warning ("in file %s: style '%s' not defined", parser_state->filename, map_to);
+	}
+
 	if (parser_state->error == NULL)
 	{
 
@@ -1571,7 +1653,7 @@ file_parse (gchar                     *filename,
 	parser_state = parser_state_new (language, ctx_data,
 					 defined_regexes, styles,
 					 replacements, reader,
-					 loaded_lang_ids);
+					 filename, loaded_lang_ids);
 	xmlTextReaderSetStructuredErrorHandler (reader,
 						(xmlStructuredErrorFunc) text_reader_structured_error_func,
 						parser_state);
@@ -1629,6 +1711,7 @@ parser_state_new (GtkSourceLanguage       *language,
 		  GHashTable              *styles_mapping,
 		  GQueue                  *replacements,
 		  xmlTextReader	          *reader,
+		  const char              *filename,
 		  GHashTable              *loaded_lang_ids)
 {
 	ParserState *parser_state;
@@ -1646,6 +1729,7 @@ parser_state_new (GtkSourceLanguage       *language,
 	parser_state->regex_compile_flags = 0;
 
 	parser_state->reader = reader;
+	parser_state->filename = g_filename_display_name (filename);
 	parser_state->error = NULL;
 
 	parser_state->defined_regexes = defined_regexes;
@@ -1678,6 +1762,7 @@ parser_state_destroy (ParserState *parser_state)
 	g_free (parser_state->closing_delimiter);
 
 	g_free (parser_state->language_decoration);
+	g_free (parser_state->filename);
 
 	g_slice_free (ParserState, parser_state);
 }
